function [epc_im] = read_epc(filename,tag,place_in_2048)
% function [epc_im] = read_epc(filename,tag,place_in_2048);
%
% Read EPIC .epc data file. Does not rotate or shuffle the image.
%
% Input:
%   filename
%       Auto-appends '.epc' to data filename if needed.
%
% optional:
%   tag     text description of file
%   place_in_2048 [0] flag, if file is sub-image, place into 2048 x 2048 image
%
% Output Structure epc_im

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Operates on EPIC .epc files, written by EPICgse. Presently only reads
% version 1.0, 2.0 and 3.0 files.
%
% 11/5/2009 Jeff Tobin, initial release
% 11/14/2009 moved image shuffles to fix_oversample.m
% 11/15/2009 added tag to argument list, and field epc_im.label
% 4/7/2010 fixed for sub-images (was bug in GSE code that made decimation
%   be complete image, rather than 512 x 512... Probably need to work what
%   happens with decimated or average data, logic not figured out yet.
% 5/13/2010 implement reader for version 1.0 and 2.0 files (july18 focus data)
%   and added place_in_2048 argument; updated reader_version to 2.0
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% configure return items so matlab won't complain for early return
epc_im.filename = '';
epc_im.struc_type ='raw_epc';
epc_im.reader_version=2.0;

% must have at least the filename argument
if nargin < 1, help(mfilename); error(mfilename); end

if nargin < 2,
  tag='';
end

if nargin < 3,
  place_in_2048=0;
end

% Build filename as necessary
[pathstr, name, ext] = fileparts(filename);
if isempty(ext),
    epc_filename=[filename '.epc'];
else
    if strcmp(ext,'.epc'),
        epc_filename=filename;
    else
        epc_filename=[filename '.epc'];
    end
end

if ~exist(epc_filename,'file') 
  error('read_epc:fileNotFound','Error, %s not found\n', epc_filename)
end

[pathstr, name, ext] = fileparts(epc_filename) ;
epc_im.filename = [name ext];
epc_im.name     = name;
epc_im.ext      = ext;
if (isempty(pathstr))
  epc_im.path   = pathstr;
else
  epc_im.path     = [pathstr '/'];
end

fid = fopen(epc_filename, 'rb');
fileinfo=dir(epc_filename);

if (fid<1)
  error('read_epc:fileUnopened','Error: found, but could not open %s \n', epc_filename)
end

% Read the file
epc_im.header = fgetl(fid);
start=strfind(epc_im.header, 'Image file for EPIC');
if isempty(start),
  warning('read_epc:unknownFile','Input is not an EPIC image file: %s',...
    epc_filename);
  return
end

epc_im.version = fscanf(fid,'%f, File format Version',1);
switch epc_im.version,
  case {1 2}
    fgetl(fid);
    fscanf(fid,'Comments: ',1);
    epc_im.comments = {fgetl(fid)};
    line=fgetl(fid); % second comment line (might be blank)
    index=1;
    while ~isempty(line),
      index=index+1;
      epc_im.comments(index)={line};
      line=fgetl(fid); % next comment line
    end

    fgetl(fid); % chuck the GSE status: line
    epc_im.CCU_A_pwr = fscanf(fid,'%d, CCU A Power 1=On, 0=Off\n',1);
    epc_im.CCU_B_pwr = fscanf(fid,'%d, CCU B Power 1=On, 0=Off\n',1);
    epc_im.MEB_A_pwr = fscanf(fid,'%d, MEB A Power 1=On, 0=Off\n',1);
    epc_im.MEB_B_pwr = fscanf(fid,'%d, MEB B Power 1=On, 0=Off\n',1);
    epc_im.gse_mode = fscanf(fid,'%d, Mode 0=CCU, 1=MEB, 2=Both GSE\n',1);
    epc_im.CCU_mode = fscanf(fid,'%d, CCU Mode\n',1);
    epc_im.offset_upper = fscanf(fid,'%d, Offset Upper\n',1);
    epc_im.offset_lower = fscanf(fid,'%d, Offset Lower\n',1);
    epc_im.exposure_ind = fscanf(fid,'%d, Exposure window\n',1);
%    epc_im.left = fscanf(fid,'%d, X-start (left edge)\n',1);
%    epc_im.top = fscanf(fid,'%d, Y-start (top edge)\n',1);
    epc_im.left = 0;
    epc_im.top = 0;
    epc_im.across = fscanf(fid,'%d, X-across (binary ',1);  %whole image, left->right then top to bottom, LSB1 MSB1 LSB2 MSB2 ...)\n',1);
    epc_im.type=fscanf(fid,'%s',1);
    fgetl(fid); % chuck the rest of the line
    epc_im.down = fscanf(fid,'%d, Y-Down\n',1);
    epc_im.nums = fscanf(fid,'%d',2); %v 3.0 has 4, looks like different meanings for different files?
% perhaps this is not correct... (not exhaustively tested, 5/13/2010)
    epc_im.left = epc_im.nums(1);
    epc_im.top = epc_im.nums(2);
    if epc_im.left==2048, epc_im.left=0; end
    if epc_im.top==2048, epc_im.top=0; end

%    epc_im.num_hk = fscanf(fid,'%d, Number of HK values\n',1);
    epc_im.num_hk=0;
    epc_im.hk=[]; %fread(fid,epc_im.num_hk,'uint32');
% pre 4/7/2010 version hard coded across/down with 2048, 2048, because
%   file headers said they were smaller, but they ended up 2048x2048. It was
%   an error in the GSE code for decimated (and average?) region. Other
%   regions worked fine. Now, all should work "properly"
%     epc_im.image=reshape(fread(fid,2048*2048,'uint16'),...
%       2048,2048);
    if (fileinfo.bytes > 600000 && epc_im.across < 512),  %512x512 is a region, and is 513 kB
      fprintf('Overriding # across, down, file is "too large". Expected 513 or 33, is %d kbytes\n', fileinfo.bytes/1024);
      epc_im.across=2048;
      epc_im.down=2048;
    end
    
% Sometimes there is a byte left, sometimes not
    while (fileinfo.bytes-ftell(fid) > 2*epc_im.across*epc_im.down),
      fread(fid,1);
    end
    epc_im.image=reshape(fread(fid,epc_im.across*epc_im.down,'uint16'),...
      epc_im.across,epc_im.down);
% check for sub-image. If it is sub-image, place in correct part of whole image
    if epc_im.across ~= 2048 && place_in_2048,
      small_im=epc_im.image;
      epc_im.image=zeros(2048,2048);
      epc_im.image(epc_im.left+1:epc_im.left+epc_im.across, ...
        epc_im.top+1:epc_im.top+epc_im.down)=small_im;
    end
    
% after writing and debugging, seems like 2 is the same as 1, perhaps I
%   haven't seen a type of file with the change???
%   case 2
%     fgetl(fid);
%     fscanf(fid,'Comments: ',1);
%     epc_im.comments = {fgetl(fid)};
%     line=fgetl(fid); % second comment line (might be blank)
%     index=1;
%     while ~isempty(line),
%       index=index+1;
%       epc_im.comments(index)={line};
%       line=fgetl(fid); % next comment line
%     end
% 
%     fgetl(fid); % chuck the GSE status: line
%     epc_im.CCU_A_pwr = fscanf(fid,'%d, CCU A Power 1=On, 0=Off\n',1);
%     epc_im.CCU_B_pwr = fscanf(fid,'%d, CCU B Power 1=On, 0=Off\n',1);
%     epc_im.MEB_A_pwr = fscanf(fid,'%d, MEB A Power 1=On, 0=Off\n',1);
%     epc_im.MEB_B_pwr = fscanf(fid,'%d, MEB B Power 1=On, 0=Off\n',1);
%     epc_im.gse_mode = fscanf(fid,'%d, Mode 0=CCU, 1=MEB, 2=Both GSE\n',1);
%     epc_im.CCU_mode = fscanf(fid,'%d, CCU Mode\n',1);
%     epc_im.offset_upper = fscanf(fid,'%d, Offset Upper\n',1);
%     epc_im.offset_lower = fscanf(fid,'%d, Offset Lower\n',1);
%     epc_im.exposure_ind = fscanf(fid,'%d, Exposure window\n',1);
% %    epc_im.left = fscanf(fid,'%d, X-start (left edge)\n',1);
% %    epc_im.top = fscanf(fid,'%d, Y-start (top edge)\n',1);
%     epc_im.left = 0;
%     epc_im.top = 0;
%     epc_im.across = fscanf(fid,'%d, X-across (binary ',1);  %spot image, left->right then top to bottom, LSB1 MSB1 LSB2 MSB2 ...)\n',1);
%     epc_im.type=fscanf(fid,'%s',1);
%     fgetl(fid); % chuck the rest of the line
%     epc_im.down = fscanf(fid,'%d, Y-Down\n',1);
%     epc_im.nums = fscanf(fid,'%d',2); %v 3.0 has 4, looks like different meanings for different files?
% % perhaps this is not correct... (not exhaustively tested, 5/13/2010)
%     epc_im.left = epc_im.nums(1);
%     epc_im.top = epc_im.nums(2);
%     if epc_im.left==2048, epc_im.left=0; end
%     if epc_im.top==2048, epc_im.top=0; end
%     
%     epc_im.num_hk=0;
%     epc_im.hk=[];
% 
% % pre 4/7/2010 version hard coded across/down with 2048, 2048, because
% %   file headers said they were smaller, but they ended up 2048x2048. It was
% %   an error in the GSE code for decimated (and average?) region. Other
% %   regions worked fine. Now, all should work "properly"
% %     epc_im.image=reshape(fread(fid,2048*2048,'uint16'),...
% %       2048,2048);
%     if (fileinfo.bytes > 600000 && epc_im.across < 512),  %512x512 is a region, and is 513 kB
%       fprintf('Overriding # across, down, file is "too large". Expected 513 or 33, is %d kbytes\n', fileinfo.bytes/1024);
%       epc_im.across=2048;
%       epc_im.down=2048;
%     end
% 
% % Sometimes there is a byte left, sometimes not
%     while (fileinfo.bytes-ftell(fid) > 2*epc_im.across*epc_im.down),
%       fread(fid,1);
%     end
%     epc_im.image=reshape(fread(fid,epc_im.across*epc_im.down,'uint16'),...
%       epc_im.across,epc_im.down);
% % check for sub-image. If it is sub-image, place in correct part of whole image
%     if epc_im.across ~= 2048 && place_in_2048,
%       small_im=epc_im.image;
%       epc_im.image=zeros(2048,2048);
%       epc_im.image(epc_im.left+1:epc_im.left+epc_im.across, ...
%         epc_im.top+1:epc_im.top+epc_im.down)=small_im;
%     end
    
  case 3
    fgetl(fid);
    fscanf(fid,'Comments: ',1);
    epc_im.comments = {fgetl(fid)};
    line=fgetl(fid); % second comment line (might be blank)
    index=1;
    while ~isempty(line),
      index=index+1;
      epc_im.comments(index)={line};
      line=fgetl(fid); % next comment line
    end

    fgetl(fid); % chuck the GSE status: line
    epc_im.CCU_A_pwr = fscanf(fid,'%d, CCU A Power 1=On, 0=Off\n',1);
    epc_im.CCU_B_pwr = fscanf(fid,'%d, CCU B Power 1=On, 0=Off\n',1);
    epc_im.MEB_A_pwr = fscanf(fid,'%d, MEB A Power 1=On, 0=Off\n',1);
    epc_im.MEB_B_pwr = fscanf(fid,'%d, MEB B Power 1=On, 0=Off\n',1);
    epc_im.gse_mode = fscanf(fid,'%d, Mode 0=CCU, 1=MEB, 2=Both GSE\n',1);
    epc_im.CCU_mode = fscanf(fid,'%d, CCU Mode\n',1);
    epc_im.offset_upper = fscanf(fid,'%d, Offset Upper\n',1);
    epc_im.offset_lower = fscanf(fid,'%d, Offset Lower\n',1);
    epc_im.exposure_ind = fscanf(fid,'%d, Exposure window\n',1);
    epc_im.left = fscanf(fid,'%d, X-start (left edge)\n',1);
    epc_im.top = fscanf(fid,'%d, Y-start (top edge)\n',1);
    epc_im.across = fscanf(fid,'%d, X-across (binary left->right then top to bottom, LSB1 MSB1 LSB2 MSB2 ...)\n',1);
    epc_im.down = fscanf(fid,'%d, Y-Down\n',1);
    epc_im.nums = fscanf(fid,'%d',4);
    epc_im.num_hk = fscanf(fid,'%d, Number of HK values\n',1);
    epc_im.hk=fread(fid,epc_im.num_hk,'uint32');
% pre 4/7/2010 version hard coded across/down with 2048, 2048, because
%   file headers said they were smaller, but they ended up 2048x2048. It was
%   an error in the GSE code for decimated (and average?) region. Other
%   regions worked fine. Now, all should work "properly"
%     epc_im.image=reshape(fread(fid,2048*2048,'uint16'),...
%       2048,2048);
    if (fileinfo.bytes > 600000 && epc_im.across <= 512),  %512x512 is a region, and is 513 kB
      fprintf('Overriding # across, down, file is "too large". Expected 513 or 33, is %d kbytes\n', fileinfo.bytes/1024);
      epc_im.across=2048;
      epc_im.down=2048;
    end
    epc_im.image=reshape(fread(fid,epc_im.across*epc_im.down,'uint16'),...
      epc_im.across,epc_im.down);
% check for sub-image. If it is sub-image, place in correct part of whole image
    if epc_im.across ~= 2048 && place_in_2048,
      small_im=epc_im.image;
      epc_im.image=zeros(2048,2048);
      epc_im.image(epc_im.left+1:epc_im.left+epc_im.across, ...
        epc_im.top+1:epc_im.top+epc_im.down)=small_im;
    end
    
  otherwise
    warning('read_epc:unknownVersion','Input is an unknown version: %d',...
      epc_im.version);
    return
end

fclose(fid);

if isempty(tag),
  tag=epc_im.filename;
end

epc_im.label=tag;

%epc_im=orderfields(epc_im);

return